import time
from typing import List, Optional, Set, Dict, Any, TYPE_CHECKING

from bxutils.logging import LogRecordType
from bxutils import logging

from bxcommon import constants
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.bloxroute.blocks_short_ids_serializer import BlockShortIds
from bxcommon.messages.bloxroute.tx_service_sync_blocks_short_ids_message import TxServiceSyncBlocksShortIdsMessage
from bxcommon.messages.bloxroute.tx_service_sync_complete_message import TxServiceSyncCompleteMessage
from bxcommon.messages.bloxroute.tx_service_sync_req_message import TxServiceSyncReqMessage
from bxcommon.messages.bloxroute.tx_service_sync_txs_message import TxServiceSyncTxsMessage
from bxcommon import log_messages
from bxcommon.models.node_type import NodeType
from bxcommon.services import tx_sync_service_helpers
from bxcommon.services.transaction_service import TransactionCacheKeyType
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.utils import performance_utils

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from bxcommon.connections.internal_node_connection import InternalNodeConnection

logger = logging.get_logger(__name__)
performance_troubleshooting_logger = logging.get_logger(
    LogRecordType.PerformanceTroubleshooting, __name__
)


class TxSyncService:
    def __init__(self, conn: "InternalNodeConnection") -> None:
        self.conn = conn
        self.node = conn.node
        self._sync_alarms: Dict[int, Any] = dict()

    def msg_tx_service_sync_txs(self, msg: TxServiceSyncTxsMessage) -> None:
        """
        Transaction service sync message receive txs data
        """

        network_num = msg.network_num()
        self.node.last_sync_message_received_by_network[network_num] = time.time()
        tx_service = self.node.get_tx_service(network_num)

        result_items = tx_service.process_tx_sync_message(msg)
        sync_metrics = self.node.sync_metrics[network_num]
        sync_metrics["msgs"] += 1
        for item in result_items:
            sync_metrics["tx_count"] += 1

            if item.content_length > 0:
                sync_metrics["tx_content_count"] += 1

            for short_id, transaction_flag in zip(
                item.short_ids, item.transaction_flag_types
            ):
                self.node.sync_short_id_buckets[network_num].incr_short_id(short_id)
                if TransactionFlag.PAID_TX in transaction_flag:
                    tx_service.set_short_id_transaction_type(short_id, transaction_flag)

        self.conn.log_debug(
            "TxSync processed msg from {} network {}, msgs: {}, transactions: {}, content: {}",
            self.conn,
            network_num,
            sync_metrics["msgs"],
            sync_metrics["tx_count"],
            sync_metrics["tx_content_count"],
        )

    def send_tx_service_sync_req(self, network_num: int):
        """
        sending transaction service sync request
        """
        self.node.last_sync_message_received_by_network[network_num] = time.time()
        self.node.sync_short_id_buckets.pop(network_num, None)
        self.node.sync_metrics.pop(network_num, None)
        self.conn.enqueue_msg(TxServiceSyncReqMessage(network_num))

        if self.node.check_sync_relay_connections_alarm_id:
            self.node.alarm_queue.unregister_alarm(self.node.check_sync_relay_connections_alarm_id)
            self.node.check_sync_relay_connections_alarm_id = None
        self.node.check_sync_relay_connections_alarm_id = self.node.alarm_queue.register_alarm(
            constants.LAST_MSG_FROM_RELAY_THRESHOLD_S, self.node.check_sync_relay_connections, self.conn
        )

    def send_tx_service_sync_complete(self, network_num: int) -> None:
        self.conn.update_tx_sync_complete(network_num)
        self.conn.enqueue_msg(TxServiceSyncCompleteMessage(network_num))

    def send_tx_service_sync_blocks_short_ids(self, network_num: int) -> None:
        blocks_short_ids: List[BlockShortIds] = []
        start_time = time.time()
        for block_hash, short_ids in self.node.get_tx_service(
            network_num
        ).iter_short_ids_seen_in_block():
            blocks_short_ids.append(BlockShortIds(block_hash, short_ids))

        block_short_ids_msg = TxServiceSyncBlocksShortIdsMessage(
            network_num, blocks_short_ids
        )
        performance_utils.log_operation_duration(
            performance_troubleshooting_logger,
            "Create tx service sync block short ids message",
            start_time,
            constants.RESPONSIVENESS_CHECK_DELAY_WARN_THRESHOLD_S,
            network_num=network_num,
            connection=self.conn,
            blocks_short_ids_count=len(blocks_short_ids),
        )
        duration = time.time() - start_time
        self.conn.log_trace(
            "Sending {} block short ids took {:.3f} seconds.",
            len(blocks_short_ids),
            duration,
        )
        self.conn.enqueue_msg(block_short_ids_msg)

    def send_tx_service_sync_txs(
        self,
        network_num: int,
        tx_service_snap: List[Sha256Hash],
        sync_tx_content: bool = True,
        duration: float = 0,
        msgs_count: int = 0,
        total_tx_count: int = 0,
        sending_tx_msgs_start_time: float = 0,
    ) -> None:
        if network_num in self._sync_alarms:
            del self._sync_alarms[network_num]

        if not self.conn.is_active():
            self.conn.log_info(
                "TxSync on network {}, sent {} transactions, and {} messages, took {:.3f}s. "
                "{} transactions were not synced. Connection had been closed",
                network_num,
                total_tx_count,
                msgs_count,
                duration,
                len(tx_service_snap),
            )
            self.send_tx_service_sync_complete(network_num)
            return

        if sending_tx_msgs_start_time == 0:
            sending_tx_msgs_start_time = time.time()
        tx_service = self.node.get_tx_service(network_num)
        sync_ping_latency = self.conn.sync_ping_latencies.get(network_num, 0.0)
        if (
            time.time() - sending_tx_msgs_start_time
        ) < constants.SENDING_TX_MSGS_TIMEOUT_S:
            if tx_service_snap and sync_ping_latency is not None:
                start = time.time()
                txs_content_short_ids = tx_sync_service_helpers.create_txs_service_msg(
                    tx_service, tx_service_snap, sync_tx_content
                )
                performance_utils.log_operation_duration(
                    performance_troubleshooting_logger,
                    "Create tx service sync message",
                    start,
                    constants.RESPONSIVENESS_CHECK_DELAY_WARN_THRESHOLD_S,
                    network_num=network_num,
                    connection=self,
                    tx_count=len(txs_content_short_ids),
                )
                self.conn.enqueue_msg(
                    TxServiceSyncTxsMessage(network_num, txs_content_short_ids)
                )
                self.conn.check_ping_latency_for_network(network_num)
                duration += time.time() - start
                msgs_count += 1
                total_tx_count += len(txs_content_short_ids)
            # checks again if tx_snap in case we still have msgs to send, else no need to wait
            # for the next interval.
            if tx_service_snap:
                if sync_ping_latency is None:
                    next_interval = constants.TX_SERVICE_SYNC_TXS_S
                else:
                    next_interval = max(
                        sync_ping_latency * 0.5, constants.TX_SERVICE_SYNC_TXS_S
                    )
                self._sync_alarms[network_num] = self.node.alarm_queue.register_alarm(
                    next_interval,
                    self.send_tx_service_sync_txs,
                    network_num,
                    tx_service_snap,
                    sync_tx_content,
                    duration,
                    msgs_count,
                    total_tx_count,
                    sending_tx_msgs_start_time,
                )
            else:  # if all txs were sent, send complete msg
                self.conn.log_debug(
                    "TxSync on network {}, sent {} transactions, and {} messages, took {:.3f}s.",
                    network_num,
                    total_tx_count,
                    msgs_count,
                    duration,
                )
                self.send_tx_service_sync_complete(network_num)
        else:  # if time is up - upgrade this node as synced - giving up
            self.conn.log_debug(
                "TxSync on network {}, sent {} transactions, and {} messages, took more than {}s. Giving up. "
                "{} transactions were not synced",
                network_num,
                total_tx_count,
                msgs_count,
                constants.SENDING_TX_MSGS_TIMEOUT_S,
                len(tx_service_snap),
            )
            self.send_tx_service_sync_complete(network_num)

    def send_tx_service_sync_txs_from_time(
        self,
        network_num: int,
        sync_tx_content: bool = True,
        duration: float = 0,
        msgs_count: int = 0,
        total_tx_count: int = 0,
        sending_tx_msgs_start_time: float = 0,
        start_time: float = 0,
        snapshot_cache_keys: Optional[Set[TransactionCacheKeyType]] = None,
    ) -> None:
        if network_num in self._sync_alarms:
            del self._sync_alarms[network_num]

        if not self.conn.is_active():
            self.conn.log_info(
                "TxSync on network {}, sent {} transactions, and {} messages, took {:.3f}s. "
                "Connection had been closed",
                self,
                network_num,
                total_tx_count,
                msgs_count,
                duration,
            )
            self.send_tx_service_sync_complete(network_num)
            return

        if sending_tx_msgs_start_time == 0:
            sending_tx_msgs_start_time = time.time()
        done = False
        last_tx_timestamp = start_time
        sync_ping_latency = self.conn.sync_ping_latencies.get(network_num, 0.0)
        tx_service = self.node.get_tx_service(network_num)
        if (
            time.time() - sending_tx_msgs_start_time
        ) < constants.SENDING_TX_MSGS_TIMEOUT_S:
            if sync_ping_latency is not None:
                start = time.time()
                (
                    txs_content_short_ids,
                    last_tx_timestamp,
                    done,
                    snapshot_cache_keys,
                ) = tx_sync_service_helpers.create_txs_service_msg_from_time(
                    tx_service, start_time, sync_tx_content, snapshot_cache_keys
                )
                self.conn.log_info(
                    "TxSync on network {}, syncing {} transactions created between {} and {} start {} end {}, "
                    "took {:.3f}s. {} transactions were synced.",
                    network_num,
                    len(snapshot_cache_keys),
                    start_time,
                    last_tx_timestamp,
                    time.time() - start,
                    "All" if done else "Not all",
                )
                performance_utils.log_operation_duration(
                    performance_troubleshooting_logger,
                    "Create tx service sync message from time",
                    start,
                    constants.RESPONSIVENESS_CHECK_DELAY_WARN_THRESHOLD_S,
                    network_num=network_num,
                    connection=self,
                    tx_count=len(txs_content_short_ids),
                )
                self.conn.enqueue_msg(
                    TxServiceSyncTxsMessage(network_num, txs_content_short_ids)
                )
                self.conn.check_ping_latency_for_network(network_num)
                duration += time.time() - start
                msgs_count += 1
                total_tx_count += len(txs_content_short_ids)
            # checks again if tx_snap in case we still have msgs to send, else no need to wait
            # for the next interval.
            if not done:
                if sync_ping_latency is None:
                    next_interval = constants.TX_SERVICE_SYNC_TXS_S
                else:
                    next_interval = max(
                        sync_ping_latency * 0.5, constants.TX_SERVICE_SYNC_TXS_S
                    )
                self._sync_alarms[network_num] = self.node.alarm_queue.register_alarm(
                    next_interval,
                    self.send_tx_service_sync_txs_from_time,
                    network_num,
                    sync_tx_content,
                    duration,
                    msgs_count,
                    total_tx_count,
                    sending_tx_msgs_start_time,
                    last_tx_timestamp,
                    snapshot_cache_keys,
                )
            else:  # if all txs were sent, send complete msg
                self.conn.log_info(
                    "TxSync on network {}, sent {} transactions, and {} messages, took {:.3f}s.",
                    network_num,
                    total_tx_count,
                    msgs_count,
                    duration,
                )
                self.send_tx_service_sync_complete(network_num)
        else:  # if time is up - upgrade this node as synced - giving up
            self.conn.log_info(
                "TxSync on network {}, sent {} transactions, and {} messages, took more than {:.3f}s. Giving up. "
                "{} transactions since were not synced",
                network_num,
                total_tx_count,
                msgs_count,
                constants.SENDING_TX_MSGS_TIMEOUT_S,
                last_tx_timestamp,
            )
            self.send_tx_service_sync_complete(network_num)

    def send_tx_service_sync_txs_from_buffer(
        self,
        network_num: int,
        txs_buffer: memoryview,
        duration: float = 0,
        msgs_count: int = 0,
        total_tx_count: int = 0,
        sending_tx_msgs_start_time: float = 0,
        start_offset: int = 0,
    ) -> None:
        if network_num in self._sync_alarms:
            del self._sync_alarms[network_num]
        if not self.conn:
            logger.warning(log_messages.CONNECTION_DOES_NOT_EXIST, "sync alarm (send_tx_service_sync_txs_from_buffer)")
            return
        if not self.conn.is_active():
            self.conn.log_info(
                "TxSync on network {}, sent {} transactions, and {} messages, took {:.3f}s. "
                "Connection had been closed",
                network_num,
                total_tx_count,
                msgs_count,
                duration,
            )
            self.send_tx_service_sync_complete(network_num)
            return

        if sending_tx_msgs_start_time == 0:
            sending_tx_msgs_start_time = time.time()

        done = False
        end_offset = start_offset

        sync_ping_latency = self.conn.sync_ping_latencies.get(network_num, 0.0)
        tx_service = self.node.get_tx_service(network_num)
        if (
            time.time() - sending_tx_msgs_start_time
        ) < constants.SENDING_TX_MSGS_TIMEOUT_S:
            if sync_ping_latency is not None:
                start = time.time()

                (
                    txs_msg,
                    txs_count,
                    end_offset,
                    done,
                ) = tx_sync_service_helpers.create_txs_service_msg_from_buffer(
                    tx_service, txs_buffer, start_offset
                )
                self.conn.log_info(
                    "TxSync on network {}, syncing {} transactions, took {:.3f}s. "
                    "Starting offset {}, ending offset {}, total offset {}. "
                    "{} transactions were synced.",
                    network_num,
                    txs_count,
                    time.time() - start,
                    start_offset,
                    end_offset,
                    len(txs_buffer),
                    "All" if done else "Not all",
                )
                performance_utils.log_operation_duration(
                    performance_troubleshooting_logger,
                    "Create tx service sync message from time",
                    start,
                    constants.RESPONSIVENESS_CHECK_DELAY_WARN_THRESHOLD_S,
                    network_num=network_num,
                    connection=self,
                    tx_count=txs_count,
                )
                self.conn.enqueue_msg(txs_msg)
                self.conn.check_ping_latency_for_network(network_num)
                duration += time.time() - start
                msgs_count += 1
                total_tx_count += txs_count

            # checks again if tx_snap in case we still have msgs to send, else no need to wait
            # for the next interval.
            if not done:
                if sync_ping_latency is None:
                    next_interval = constants.TX_SERVICE_SYNC_TXS_S
                else:
                    next_interval = max(
                        sync_ping_latency * 0.5, constants.TX_SERVICE_SYNC_TXS_S
                    )
                self._sync_alarms[network_num] = self.node.alarm_queue.register_alarm(
                    next_interval,
                    self.send_tx_service_sync_txs_from_buffer,
                    network_num,
                    txs_buffer,
                    duration,
                    msgs_count,
                    total_tx_count,
                    sending_tx_msgs_start_time,
                    end_offset,
                )
            else:  # if all txs were sent, send complete msg
                self.conn.log_info(
                    "TxSync on network {}, sent {} transactions, and {} messages. took {:.3f}s.",
                    network_num,
                    total_tx_count,
                    msgs_count,
                    duration,
                )
                self.send_tx_service_sync_complete(network_num)
        else:  # if time is up - upgrade this node as synced - giving up
            self.conn.log_info(
                "TxSync on network {}, sent {} transactions, and {} messages, took more than {:.3f}s. Giving up. "
                "{} bytes out of {} total were synced",
                network_num,
                total_tx_count,
                msgs_count,
                constants.SENDING_TX_MSGS_TIMEOUT_S,
                start_offset,
                len(txs_buffer),
            )
            self.send_tx_service_sync_complete(network_num)

    def msg_tx_service_sync_complete(self, msg: TxServiceSyncCompleteMessage) -> None:
        if (
            self.node.NODE_TYPE in NodeType.GATEWAY
            # pylint: disable=unsupported-membership-test
            and ConnectionType.RELAY_BLOCK == self.conn.CONNECTION_TYPE
        ):
            return
        if self.node.check_sync_relay_connections_alarm_id:
            self.node.alarm_queue.unregister_alarm(self.node.check_sync_relay_connections_alarm_id)
            self.node.check_sync_relay_connections_alarm_id = None
        if self.node.transaction_sync_timeout_alarm_id:
            self.node.alarm_queue.unregister_alarm(self.node.transaction_sync_timeout_alarm_id)
            self.node.transaction_sync_timeout_alarm_id = None
        network_num = msg.network_num()
        self.node.on_network_synced(network_num)
        duration = time.time() - self.node.start_sync_time
        self.conn.log_info(
            "TxSync complete. It took {:.3f} seconds to complete transaction state with BDN.",
            duration,
        )
        sync_data: Dict[str, Any] = {"peer_id": self.conn.peer_id, "duration": duration}
        for network_num, sync_metrics in self.node.sync_metrics.items():
            tx_service = self.node.get_tx_service(network_num)
            network_stats = dict(sync_metrics)
            network_stats["content_without_sid"] = len(
                tx_service.tx_hashes_without_short_id.queue
            )
            network_stats["sid_without_content"] = len(
                tx_service.tx_hashes_without_content.queue
            )
            # pylint: disable=protected-access
            # noinspection PyProtectedMember
            network_stats["tx_content_len"] = len(tx_service._tx_cache_key_to_contents)
            sync_data[network_num] = network_stats

        sync_data["short_id_buckets"] = self.node.sync_short_id_buckets

        logger.debug({"type": "TxSyncMetrics", "data": sync_data})
        self.node.on_fully_updated_tx_service()

    def dispose(self):
        self.conn = None



import blxr_rlp as rlp

from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType


class StatusEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.STATUS

    fields = [("eth_version", rlp.sedes.big_endian_int),
              ("network_id", rlp.sedes.big_endian_int),
              ("chain_difficulty", rlp.sedes.big_endian_int),
              ("chain_head_hash", rlp.sedes.binary),
              ("genesis_hash", rlp.sedes.binary),
              ("fork_id", rlp.sedes.List([rlp.sedes.binary, rlp.sedes.big_endian_int], strict=False))]

    def get_eth_version(self):
        return self.get_field_value("eth_version")

    def get_network_id(self):
        return self.get_field_value("network_id")

    def get_chain_difficulty(self):
        return self.get_field_value("chain_difficulty")

    def get_chain_head_hash(self):
        return self.get_field_value("chain_head_hash")

    def get_genesis_hash(self):
        return self.get_field_value("genesis_hash")

    def get_fork_id(self):
        return self.get_field_value("fork_id")



import asyncio
import os
from typing import Optional, List, TYPE_CHECKING
from bxcommon.feed.feed_manager import FeedManager
from bxgateway.rpc.ws.ws_connection import WsConnection
from bxgateway.rpc.subscription_rpc_handler import SubscriptionRpcHandler
from bxutils import logging
from bxcommon.utils import config

import websockets
from websockets import WebSocketServerProtocol
from websockets.server import WebSocketServer

from bxutils.encoding.json_encoder import Case

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

logger = logging.get_logger(__name__)


class IpcServer:
    def __init__(
        self,
        ipc_file: str,
        feed_manager: FeedManager,
        node: "AbstractGatewayNode",
        case: Case = Case.CAMEL
    ):
        self.ipc_path = config.get_data_file(ipc_file)
        self.node = node
        self.feed_manager = feed_manager
        self.case = case
        self._server: Optional[WebSocketServer] = None
        self._connections: List[WsConnection] = []
        self._started: bool = False

    def status(self) -> bool:
        return self._started

    async def start(self) -> None:
        if os.path.exists(self.ipc_path):
            os.remove(self.ipc_path)
        self._server = await websockets.unix_serve(self.handle_connection, self.ipc_path)
        self._started = True

    async def stop(self) -> None:
        server = self._server
        self._started = False
        if server is not None:
            await asyncio.gather(
                *(connection.close() for connection in self._connections)
            )

            server.close()
            await server.wait_closed()
        if os.path.exists(self.ipc_path):
            os.remove(self.ipc_path)

    async def handle_connection(self, websocket: WebSocketServerProtocol, path: str) -> None:
        logger.trace("Accepting new IPC connection...")
        connection = WsConnection(
            websocket,
            path,
            SubscriptionRpcHandler(self.node, self.feed_manager, self.case)
        )
        self._connections.append(connection)
        await connection.handle()
        self._connections.remove(connection)



from bxcommon.models.notification_code import NotificationCode

# notifications are being formatted according to the provided format string
# the formatting arguments (list) are as follows
#

# 0    - notification code number
# 1    - notification code label
# 2..n -  notification content items in order, content is treated as csv
default_notification_format = "{1}({0}): {2}"

# {percentage}% of daily {entity_type} quota with limit of {limit} {entity_type}s per day is depleted.
quota_notification_format = "{0}% of daily {1} quota with limit of {2} {1}s per day is depleted."
account_expired_notification_format = "The account associated with this gateway has expired. " \
                                      "Please visit https://portal.bloxroute.com to renew your subscription."

NotificationFormatting = {
    NotificationCode.QUOTA_FILL_STATUS: quota_notification_format,
    NotificationCode.ACCOUNT_EXPIRED_NOTIFICATION: account_expired_notification_format
}



from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType


class ConfirmedTxMessage(AbstractBloxrouteMessage):
    """
    Confirmation message that a transaction has been accepted into the
    blockchain node's mempool. Only sent to/from streaming gateways.
    """

    MESSAGE_TYPE = GatewayMessageType.CONFIRMED_TX
    MSG_SIZE = (
        AbstractBloxrouteMessage.HEADER_LENGTH
        + crypto.SHA256_HASH_LEN
        + constants.CONTROL_FLAGS_LEN
    )
    PAYLOAD_LENGTH = crypto.SHA256_HASH_LEN + constants.CONTROL_FLAGS_LEN
    EMPTY_TX_VAL = TxMessage.EMPTY_TX_VAL

    buf: bytearray
    _tx_hash: Optional[Sha256Hash] = None
    _tx_val: Optional[memoryview] = None

    def __init__(
        self,
        tx_hash: Optional[Sha256Hash] = None,
        tx_val: memoryview = TxMessage.EMPTY_TX_VAL,
        buf: Optional[bytearray] = None,
    ) -> None:
        if buf is None:
            assert tx_hash is not None
            assert tx_val is not None
            buf = bytearray(self.MSG_SIZE + len(tx_val))

            off = AbstractBloxrouteMessage.HEADER_LENGTH
            buf[off:off + crypto.SHA256_HASH_LEN] = tx_hash.binary

            off += crypto.SHA256_HASH_LEN
            buf[off:off + len(tx_val)] = tx_val

        self.buf = buf
        super().__init__(self.MESSAGE_TYPE, self.PAYLOAD_LENGTH + len(tx_val), buf)

    def tx_hash(self) -> Sha256Hash:
        if self._tx_hash is None:
            off = AbstractBloxrouteMessage.HEADER_LENGTH
            self._tx_hash = Sha256Hash(self._memoryview[off:off + crypto.SHA256_HASH_LEN])

        tx_hash = self._tx_hash
        assert tx_hash is not None
        return tx_hash

    def tx_val(self) -> memoryview:
        if self._tx_val is None:
            off = AbstractBloxrouteMessage.HEADER_LENGTH + crypto.SHA256_HASH_LEN
            if len(self.buf) == self.PAYLOAD_LENGTH:
                self._tx_val = ConfirmedTxMessage.EMPTY_TX_VAL
            else:
                self._tx_val = self._memoryview[off:-constants.CONTROL_FLAGS_LEN]

        tx_val = self._tx_val
        assert tx_val is not None
        return tx_val



from bxcommon.models.serializeable_enum import SerializeableEnum


class RpcRequestType(SerializeableEnum):
    BLXR_TX = 0
    GATEWAY_STATUS = 1
    STOP = 2
    MEMORY = 3
    PEERS = 4
    BDN_PERFORMANCE = 5
    HEALTHCHECK = 6
    PING = 7
    SUBSCRIBE = 8
    UNSUBSCRIBE = 9
    QUOTA_USAGE = 10
    MEMORY_USAGE = 11
    TX_STATUS = 12
    BLXR_ETH_CALL = 13
    TX_SERVICE = 14
    ADD_BLOCKCHAIN_PEER = 15
    REMOVE_BLOCKCHAIN_PEER = 16
    _START_MONITOR_TRANSACTION_BY_HASH = 17
    STOP_MONITOR_TRANSACTION = 18
    LIST_ALL_MONITORED_TRANSACTIONS = 19
    BLXR_PRIVATE_TX = 20
    START_MONITOR_TRANSACTION = 21
    START_TRANSACTION_FEE_BUMP = 22
    STOP_TRANSACTION_FEE_BUMP = 23
    BLXR_TX_FEE_BUMP = 24
    NETWORK_TX_FEE_UPDATE = 25
    BLXR_SIMULATE_BUNDLE = 26
    BLXR_SUBMIT_BUNDLE = 27
    BLXR_PROFIT_SHARING_PRIVATE_TX = 28
    METAMASK_PRIVATE_TX = 29
    SIMULATE_ARB_ONLY_BUNDLE = 30
    SUBMIT_ARB_ONLY_BUNDLE = 31
    BLXR_INFO = 32
    BACKRUN_PRIVATE_TX = 33
    ADD_MEV_CREDIT = 34
    GET_MEV_CREDIT = 35



import struct
from typing import Optional, List

from bxcommon.constants import UL_INT_SIZE_IN_BYTES, CONTROL_FLAGS_LEN, UL_SHORT_SIZE_IN_BYTES
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.v6 import txs_serializer_v6
from bxcommon.messages.bloxroute.v6.txs_serializer_v6 import TxContentShortIdsV6
from bxutils.logging.log_level import LogLevel


class TxServiceSyncTxsMessageV6(AbstractBloxrouteMessage):
    MESSAGE_TYPE = BloxrouteMessageType.TX_SERVICE_SYNC_TXS
    """
    Message used to send information about txs hash from requested relay
    """

    def __init__(
            self, network_num: Optional[int] = None,
            txs_content_short_ids: Optional[List[TxContentShortIdsV6]] = None,
            buf: Optional[bytearray] = None
    ):
        # pyre-fixme[8]: Attribute has type `int`; used as `Optional[int]`.
        self._network_num: int = network_num
        # pyre-fixme[8]: Attribute has type `int`; used as `Optional[int]`.
        self._tx_count: int = len(txs_content_short_ids) if txs_content_short_ids is not None else None
        # pyre-fixme[8]: Attribute has type `List[TxContentShortIdsV6]`; used as
        #  `Optional[List[TxContentShortIdsV6]]`.
        self._txs_content_short_ids: List[TxContentShortIdsV6] = txs_content_short_ids

        if txs_content_short_ids is not None and buf is None:
            self.buf = bytearray(self.HEADER_LENGTH + UL_INT_SIZE_IN_BYTES + UL_INT_SIZE_IN_BYTES)
            self._parse()
        elif buf is not None:
            self.buf = buf

        self.buf.extend(bytearray(CONTROL_FLAGS_LEN))
        super(TxServiceSyncTxsMessageV6, self).__init__(
            self.MESSAGE_TYPE,
            len(self.buf) - self.HEADER_LENGTH,
            self.buf
        )

    def log_level(self):
        return LogLevel.DEBUG

    def network_num(self) -> int:
        if self._network_num is None:
            off = self.HEADER_LENGTH
            self._network_num, = struct.unpack_from("<L", self._memoryview, off)
        return self._network_num

    def tx_count(self) -> int:
        if self._tx_count is None:
            off = self.HEADER_LENGTH + UL_INT_SIZE_IN_BYTES
            self._tx_count, = struct.unpack_from("<L", self._memoryview, off)
        return self._tx_count

    def txs_content_short_ids(self) -> List[TxContentShortIdsV6]:
        offset = self.HEADER_LENGTH + UL_INT_SIZE_IN_BYTES + UL_INT_SIZE_IN_BYTES
        return txs_serializer_v6.deserialize_txs_content_short_ids_from_buffer(
            self._memoryview, offset, self.tx_count()
        )

    def __repr__(self) -> str:
        return "{}<network_num: {}, tx_count: {}".format(self.__class__.__name__, self.network_num(), self.tx_count())

    def _parse(self) -> None:
        off = self.HEADER_LENGTH
        tx_count = len(self._txs_content_short_ids)
        struct.pack_into("<LL", self.buf, off, self._network_num, tx_count)
        off += UL_INT_SIZE_IN_BYTES + UL_SHORT_SIZE_IN_BYTES

        self.buf.extend(
            txs_serializer_v6.serialize_txs_content_short_ids_into_bytes(
                self._txs_content_short_ids, self.network_num()
            )
        )



from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxutils.logging import LogLevel


class PingEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.PING



from aiohttp.abc import StreamResponse
from aiohttp.web import Response


class ResponseFormatter:
    _response: Response

    def __init__(self, response: StreamResponse) -> None:
        assert isinstance(response, Response)
        self._response = response

    def __repr__(self) -> str:
        response = self._response
        return f"HTTPResponse <status: {response.status}, body: {response.body}>"



import datetime
import struct
from collections import defaultdict
from typing import Optional, Dict, List, TYPE_CHECKING

from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

from bxcommon import constants
from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats.stat_event_type_settings import StatEventTypeSettings
from bxcommon.utils.stats.statistics_event_service import StatisticsEventService

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from bxcommon.connections.abstract_connection import AbstractConnection

logger = logging.get_logger(__name__)


class _TransactionStatisticsService(StatisticsEventService):
    def __init__(self) -> None:
        super(_TransactionStatisticsService, self).__init__()
        self.name = "TransactionInfo"
        self.logger = logging.get_logger(LogRecordType.TransactionInfo)
        self.priority_logger = logging.get_logger(LogRecordType.TransactionPropagationInfo)

        self.log_percentage_for_hash_by_network_num: Dict[int, float] = \
            defaultdict(lambda: constants.TRANSACTIONS_BY_HASH_PERCENTAGE_TO_LOG_STATS_FOR)
        self.log_percentage_for_sid_by_network_num: Dict[int, float] = \
            defaultdict(lambda: constants.TRANSACTIONS_BY_SID_PERCENTAGE_TO_LOG_STATS_FOR)

    def configure_network(self, network_num: int, percent_to_log_by_hash: float, percent_to_log_by_sid: float) -> None:
        self.log_percentage_for_hash_by_network_num[network_num] = percent_to_log_by_hash
        self.log_percentage_for_sid_by_network_num[network_num] = percent_to_log_by_sid

    def add_tx_by_hash_event(
        self,
        tx_hash: Sha256Hash,
        tx_event_settings: StatEventTypeSettings,
        network_num: int,
        short_id: Optional[int] = None,
        start_date_time: Optional[datetime.datetime] = None,
        end_date_time: Optional[datetime.datetime] = None,
        peers: Optional[List["AbstractConnection"]] = None,
        **kwargs
    ) -> None:
        tx_hash = tx_hash.binary
        if self.should_log_event_for_tx(tx_hash, network_num, short_id):
            self.log_event(tx_event_settings, convert.bytes_to_hex(tx_hash), start_date_time, end_date_time,
                           short_id=short_id, network_num=network_num, peers=peers, **kwargs)

    def add_txs_by_short_ids_event(
        self,
        short_ids,
        tx_event_settings: StatEventTypeSettings,
        network_num: int,
        start_date_time: Optional[datetime.datetime] = None,
        end_date_time: Optional[datetime.datetime] = None,
        peers: Optional[List["AbstractConnection"]] = None,
        **kwargs
    ) -> None:
        if not constants.ENABLE_TRANSACTIONS_STATS_BY_SHORT_IDS:
            return

        if len(short_ids) < 1:
            # TODO: should be an assertion
            logger.debug("Attempted to log message with 0 short ids!")
            return

        if not tx_event_settings:
            raise ValueError("tx_event_name is required")

        if self.log_percentage_for_hash_by_network_num[network_num] >= 0:
            self.log_event(
                tx_event_settings, short_ids, start_date_time, end_date_time,
                network_num=network_num, peers=peers, **kwargs
            )

    def should_log_event_for_tx(
        self, tx_hash_bytes: bytearray, network_num: int, short_id: Optional[int]
    ) -> bool:
        percent_to_log_by_tx_hash = self.log_percentage_for_hash_by_network_num[network_num]
        percent_to_log_by_sid = self.log_percentage_for_sid_by_network_num[network_num]
        if percent_to_log_by_tx_hash <= 0 and percent_to_log_by_sid <= 0:
            return False

        last_bytes_value, = struct.unpack_from(
            ">H",
            tx_hash_bytes, len(tx_hash_bytes) - constants.TAIL_BYTE_COUNT
        )
        log_tx_stat_probability_value = float(last_bytes_value) * 100 / constants.MAX_TAIL_BYTE_VALUE
        should_log_tx_hash = log_tx_stat_probability_value <= percent_to_log_by_tx_hash

        should_log_short_id = False
        # exclude short_id == 0 as well
        if short_id:
            should_log_short_id = float(short_id % 10000 + 1) / 100 <= percent_to_log_by_sid

        return should_log_tx_hash or should_log_short_id


tx_stats = _TransactionStatisticsService()



import blxr_rlp as rlp
from bxcommon.utils.blockchain_utils.eth import eth_common_constants


class BlockHash(rlp.Serializable):
    fields = [
        ("hash", rlp.sedes.Binary.fixed_length(eth_common_constants.BLOCK_HASH_LEN)),
        ("number", rlp.sedes.big_endian_int)
    ]



class PortRange:
    start: int
    end: int

    def __init__(self, start: int, end: int) -> None:
        self.start = start
        self.end = end

    def __contains__(self, port: int) -> bool:
        return self.start <= port <= self.end

    def __repr__(self) -> str:
        return f"[{self.start}:{self.end}]"



import time
from typing import Type

from bxcommon.exceptions import ParseError
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory, MessagePreview
from bxcommon.utils.buffers.input_buffer import InputBuffer
from bxcommon.utils.blockchain_utils.eth import eth_common_constants
from bxgateway import gateway_constants
from bxgateway.messages.eth.protocol.block_bodies_eth_protocol_message import BlockBodiesEthProtocolMessage
from bxgateway.messages.eth.protocol.block_headers_eth_protocol_message import BlockHeadersEthProtocolMessage
from bxgateway.messages.eth.protocol.disconnect_eth_protocol_message import DisconnectEthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxgateway.messages.eth.protocol.get_block_bodies_eth_protocol_message import GetBlockBodiesEthProtocolMessage
from bxgateway.messages.eth.protocol.get_block_headers_eth_protocol_message import GetBlockHeadersEthProtocolMessage
from bxgateway.messages.eth.protocol.get_node_data_eth_protocol_message import GetNodeDataEthProtocolMessage
from bxgateway.messages.eth.protocol.get_pooled_transactions_eth_protocol_message import \
    GetPooledTransactionsEthProtocolMessage
from bxgateway.messages.eth.protocol.get_receipts_eth_protocol_message import GetReceiptsEthProtocolMessage
from bxgateway.messages.eth.protocol.hello_eth_protocol_message import HelloEthProtocolMessage
from bxgateway.messages.eth.protocol.new_block_eth_protocol_message import NewBlockEthProtocolMessage
from bxgateway.messages.eth.protocol.new_block_hashes_eth_protocol_message import NewBlockHashesEthProtocolMessage
from bxgateway.messages.eth.protocol.new_pooled_transaction_hashes_eth_protocol_message import \
    NewPooledTransactionHashesEthProtocolMessage
from bxgateway.messages.eth.protocol.node_data_eth_protocol_message import NodeDataEthProtocolMessage
from bxgateway.messages.eth.protocol.ping_eth_protocol_message import PingEthProtocolMessage
from bxgateway.messages.eth.protocol.pong_eth_protocol_message import PongEthProtocolMessage
from bxgateway.messages.eth.protocol.pooled_transactions_eth_protocol_message import \
    PooledTransactionsEthProtocolMessage
from bxgateway.messages.eth.protocol.raw_eth_protocol_message import RawEthProtocolMessage
from bxgateway.messages.eth.protocol.receipts_eth_protocol_message import ReceiptsEthProtocolMessage
from bxgateway.messages.eth.protocol.status_eth_protocol_message import StatusEthProtocolMessage
from bxgateway.messages.eth.protocol.status_eth_protocol_message_v63 import StatusEthProtocolMessageV63
from bxgateway.messages.eth.protocol.transactions_eth_protocol_message import TransactionsEthProtocolMessage
from bxgateway.utils.eth.framed_input_buffer import FramedInputBuffer
from bxgateway.utils.eth.rlpx_cipher import RLPxCipher
from bxgateway.utils.stats.eth.eth_gateway_stats_service import eth_gateway_stats_service


class EthProtocolMessageFactory(AbstractMessageFactory):
    _MESSAGE_TYPE_MAPPING = {
        EthProtocolMessageType.HELLO: HelloEthProtocolMessage,
        EthProtocolMessageType.DISCONNECT: DisconnectEthProtocolMessage,
        EthProtocolMessageType.PING: PingEthProtocolMessage,
        EthProtocolMessageType.PONG: PongEthProtocolMessage,
        EthProtocolMessageType.TRANSACTIONS: TransactionsEthProtocolMessage,
        EthProtocolMessageType.NEW_BLOCK_HASHES: NewBlockHashesEthProtocolMessage,
        EthProtocolMessageType.GET_BLOCK_HEADERS: GetBlockHeadersEthProtocolMessage,
        EthProtocolMessageType.BLOCK_HEADERS: BlockHeadersEthProtocolMessage,
        EthProtocolMessageType.GET_BLOCK_BODIES: GetBlockBodiesEthProtocolMessage,
        EthProtocolMessageType.BLOCK_BODIES: BlockBodiesEthProtocolMessage,
        EthProtocolMessageType.NEW_BLOCK: NewBlockEthProtocolMessage,
        EthProtocolMessageType.GET_NODE_DATA: GetNodeDataEthProtocolMessage,
        EthProtocolMessageType.NODE_DATA: NodeDataEthProtocolMessage,
        EthProtocolMessageType.GET_RECEIPTS: GetReceiptsEthProtocolMessage,
        EthProtocolMessageType.RECEIPTS: ReceiptsEthProtocolMessage,
        EthProtocolMessageType.NEW_POOLED_TRANSACTION_HASHES: NewPooledTransactionHashesEthProtocolMessage,
        EthProtocolMessageType.GET_POOLED_TRANSACTIONS: GetPooledTransactionsEthProtocolMessage,
        EthProtocolMessageType.POOLED_TRANSACTIONS: PooledTransactionsEthProtocolMessage,
    }

    if gateway_constants.ETH_PROTOCOL_VERSION_63 == eth_common_constants.ETH_PROTOCOL_VERSION:
        _MESSAGE_TYPE_MAPPING.update({EthProtocolMessageType.STATUS: StatusEthProtocolMessageV63})
    else:
        _MESSAGE_TYPE_MAPPING.update({EthProtocolMessageType.STATUS: StatusEthProtocolMessage})

    def __init__(self, rlpx_cipher):
        if not isinstance(rlpx_cipher, RLPxCipher):
            raise TypeError("Argument rlpx_cipher is expected to be of type RLPxCipher but was {}"
                            .format(type(rlpx_cipher)))

        super(EthProtocolMessageFactory, self).__init__()

        self.message_type_mapping = self._MESSAGE_TYPE_MAPPING
        self._framed_input_buffer = FramedInputBuffer(rlpx_cipher)
        self._expected_msg_type = None

    def get_base_message_type(self) -> Type[AbstractMessage]:
        return EthProtocolMessage

    def set_expected_msg_type(self, msg_type):
        if not msg_type in [EthProtocolMessageType.AUTH, EthProtocolMessageType.AUTH_ACK]:
            raise ValueError("msg_type can be AUTH or AUTH_ACK")

        self._expected_msg_type = msg_type

    def reset_expected_msg_type(self):
        self._expected_msg_type = None

    def get_message_header_preview_from_input_buffer(self, input_buffer: InputBuffer) -> MessagePreview:
        """
        Peeks at a message, determining if its full.
        Returns (is_full_message, command, payload_length)
        """
        if self._expected_msg_type == EthProtocolMessageType.AUTH:
            return MessagePreview(True, EthProtocolMessageType.AUTH, input_buffer.length)
        elif self._expected_msg_type == EthProtocolMessageType.AUTH_ACK:
            auth_ack_len = self._framed_input_buffer.peek_eip8_handshake_message_len(input_buffer)
            if auth_ack_len:
                return MessagePreview(True, EthProtocolMessageType.AUTH_ACK, auth_ack_len)
        elif self._expected_msg_type is None:
            decryption_start_time = time.time()
            is_full_msg, command = self._framed_input_buffer.peek_message(input_buffer)
            eth_gateway_stats_service.log_decrypted_message(time.time() - decryption_start_time)

            if is_full_msg:
                return MessagePreview(True, command, 0)

        return MessagePreview(False, None, None)

    def create_message_from_buffer(self, buf):
        """
        Parses a full message from a buffer based on its command into one of the loaded message types.
        """

        if self._expected_msg_type is not None:
            return self.base_message_type.initialize_class(RawEthProtocolMessage, buf, None)

        if len(buf) != 0:
            raise ParseError("All bytes are expected to be already read by self._framed_input_buffer")

        message_bytes, command = self._framed_input_buffer.get_full_message()

        return self.create_message(command, message_bytes)



import logging
import sys
from io import BytesIO

import msgpack

from bxutils import constants
from bxutils import log_messages
from bxutils.logging.fluentd_overflow_handler_type import OverflowHandlerType

logger = logging.getLogger(__name__)


def overflow_handler_print(pending_records_buffer, file_obj=sys.stdout):
    logger.error(log_messages.FLUENTD_LOGGER_BUFFER_OVERFLOW)
    unpacker = msgpack.Unpacker(BytesIO(pending_records_buffer))
    for unpacked in unpacker:
        print(unpacked, file=file_obj)


def overflow_handler_ignore(_pending_records_buffer):
    logger.error(log_messages.FLUENTD_LOGGER_BUFFER_OVERFLOW)


overflow_handler_map = {
    OverflowHandlerType.Ignore: overflow_handler_ignore,
    OverflowHandlerType.Print: overflow_handler_print,
}

overflow_handler = overflow_handler_map[constants.FLUENTD_OVERFLOW_HANDLER]



from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash
from bxcommon.utils.blockchain_utils.ont.ont_common_constants import ONT_HASH_LEN


class OntObjectHash(BtcObjectHash):
    def __repr__(self):
        return "OntObjectHash<binary: {}>".format(convert.bytes_to_hex(self.binary))


NULL_ONT_BLOCK_HASH = OntObjectHash(binary=bytearray(ONT_HASH_LEN))



import time
import typing
from datetime import datetime

from bxcommon.messages.bloxroute.abstract_cleanup_message import AbstractCleanupMessage
from bxcommon.services.extension_transaction_service import ExtensionTransactionService
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils.proxy import task_pool_proxy
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType


import task_pool_executor as tpe

logger = logging.get_logger(LogRecordType.TransactionCleanup, __name__)


def contents_cleanup(
    transaction_service: TransactionService,
    block_confirmation_message: AbstractCleanupMessage,
    cleanup_tasks
):
    start_datetime = datetime.utcnow()
    start_time = time.time()
    tx_service = typing.cast(ExtensionTransactionService, transaction_service)
    cleanup_task = cleanup_tasks.borrow_task()
    cleanup_task.init(tpe.InputBytes(block_confirmation_message.buf), tx_service.proxy)
    task_pool_proxy.run_task(cleanup_task)
    short_ids = cleanup_task.short_ids()
    total_content_removed = cleanup_task.total_content_removed()
    tx_count = cleanup_task.tx_count()
    message_hash = block_confirmation_message.message_hash()
    tx_service.update_removed_transactions(total_content_removed, short_ids)
    transaction_service.on_block_cleaned_up(message_hash)
    end_datetime = datetime.utcnow()
    end_time = time.time()
    duration = end_time - start_time
    logger.statistics(
        {
            "type": "MemoryCleanup",
            "event": "CacheStateAfterBlockCleanup",
            "data": transaction_service.get_cache_state_json(),
            "start_datetime": start_datetime,
            "end_datetime": end_datetime,
            "duration": duration,
            "total_content_removed": total_content_removed,
            "tx_count": tx_count,
            "short_ids_count": len(short_ids),
            "message_hash": repr(message_hash),
        }
    )
    cleanup_tasks.return_task(cleanup_task)



from typing import Type

from bxcommon.exceptions import ParseError, UnrecognizedCommandError
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory, MessagePreview
from bxcommon.utils.buffers.input_buffer import InputBuffer
from bxgateway.messages.eth.discovery.eth_discovery_message import EthDiscoveryMessage
from bxgateway.messages.eth.discovery.eth_discovery_message_type import EthDiscoveryMessageType
from bxgateway.messages.eth.discovery.ping_eth_discovery_message import PingEthDiscoveryMessage
from bxgateway.messages.eth.discovery.pong_eth_discovery_message import PongEthDiscoveryMessage
from bxcommon.utils.blockchain_utils.eth import rlp_utils, eth_common_constants


class _EthDiscoveryMessageFactory(AbstractMessageFactory):
    _MESSAGE_TYPE_MAPPING = {
        EthDiscoveryMessageType.PING: PingEthDiscoveryMessage,
        EthDiscoveryMessageType.PONG: PongEthDiscoveryMessage
    }

    def __init__(self):
        super(_EthDiscoveryMessageFactory, self).__init__()
        self.message_type_mapping = self._MESSAGE_TYPE_MAPPING

    def get_base_message_type(self) -> Type[AbstractMessage]:
        return EthDiscoveryMessage

    def get_message_header_preview_from_input_buffer(self, input_buffer: InputBuffer) -> MessagePreview:
        """
        Peeks at a message, determining if its full.
        Returns (is_full_message, command, payload_length)
        """
        # pyre-fixme[25]: Assertion will always fail.
        if not isinstance(input_buffer, InputBuffer):
            raise ValueError("InputBuffer type is expected")

        msg_type = self._peek_message_type(input_buffer)

        if msg_type is None:
            return MessagePreview(False, None, None)

        msg_len = self._peek_message_len(input_buffer)

        is_full = input_buffer.length >= msg_len

        return MessagePreview(is_full, msg_type, msg_len)

    def create_message_from_buffer(self, buf):
        """
        Parses a full message from a buffer based on its command into one of the loaded message types.
        """
        command = self._get_message_type(buf)

        if command not in self.message_type_mapping:
            raise UnrecognizedCommandError("Message not recognized: {0}. Raw data: {1}".format(command, repr(buf)), buf)

        message_cls = self.message_type_mapping[command]
        return self.base_message_type.initialize_class(message_cls, buf, None)

    def _peek_message_type(self, input_buffer):
        if not input_buffer:
            return None

        if not isinstance(input_buffer, InputBuffer):
            raise ValueError("InputBuffer type is expected")

        msg_type_position = eth_common_constants.MDC_LEN + eth_common_constants.SIGNATURE_LEN

        if input_buffer.length < msg_type_position + eth_common_constants.MSG_TYPE_LEN:
            return None

        msg_type_bytes = input_buffer.get_slice(msg_type_position, msg_type_position + eth_common_constants.MSG_TYPE_LEN)
        msg_type = rlp_utils.safe_ord(msg_type_bytes)

        return msg_type

    def _peek_message_len(self, input_buffer):
        if not input_buffer:
            return None

        if not isinstance(input_buffer, InputBuffer):
            raise ValueError("InputBuffer type is expected")

        msg_content_position = eth_common_constants.MDC_LEN + eth_common_constants.SIGNATURE_LEN + eth_common_constants.MSG_TYPE_LEN

        if input_buffer.length < msg_content_position:
            return None

        msg_content = input_buffer.get_slice(msg_content_position, input_buffer.length)
        msg_content_memview = memoryview(msg_content)

        _, content_length, length_prefix_size = rlp_utils.consume_length_prefix(msg_content_memview, 0)

        return msg_content_position + length_prefix_size + content_length

    def _get_message_type(self, msg_bytes):
        msg_type_position = eth_common_constants.MDC_LEN + eth_common_constants.SIGNATURE_LEN

        if len(msg_bytes) <= msg_type_position:
            raise ParseError("Message length {0} is less then position of message type.".format(len(msg_bytes)))

        msg_type = rlp_utils.safe_ord(msg_bytes[msg_type_position:msg_type_position + eth_common_constants.MSG_TYPE_LEN])
        return msg_type


eth_discovery_message_factory = _EthDiscoveryMessageFactory()



from bxgateway.messages.btc.btc_normal_message_converter import BtcNormalMessageConverter


def create_btc_message_converter(magic, opts):
    if opts.use_extensions or opts.import_extensions:
        from bxgateway.messages.btc.btc_extension_message_converter import BtcExtensionMessageConverter

    if opts.use_extensions:
        return BtcExtensionMessageConverter(magic)
    else:
        return BtcNormalMessageConverter(magic)



from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.v6.tx_message_converter_v6 import tx_message_converter_v6
from bxcommon.messages.bloxroute.v6.tx_sync_message_converter_v6 import tx_sync_message_converter_v6
from bxcommon.messages.bloxroute.v7.message_converter_factory_v7 import message_converter_factory_v7
from bxcommon.messages.versioning.abstract_version_converter_factory import AbstractMessageConverterFactory
from bxcommon.messages.versioning.no_changes_message_converter import no_changes_message_converter


class _MessageConverterFactoryV6(AbstractMessageConverterFactory):
    _MESSAGE_CONVERTER_MAPPING = {
        # pylint: disable=protected-access
        **message_converter_factory_v7._MESSAGE_CONVERTER_MAPPING,
        BloxrouteMessageType.TRANSACTION: tx_message_converter_v6,
        BloxrouteMessageType.TX_SERVICE_SYNC_TXS: tx_sync_message_converter_v6,
    }

    def get_message_converter(self, msg_type):
        if not msg_type:
            raise ValueError("msg_type is required.")

        if msg_type not in self._MESSAGE_CONVERTER_MAPPING:
            return no_changes_message_converter

        return self._MESSAGE_CONVERTER_MAPPING[msg_type]


message_converter_factory_v6 = _MessageConverterFactoryV6()



from typing import TYPE_CHECKING

from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.rpc.rpc_errors import RpcAccountIdError
from bxcommon.services import sdn_http_service

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class QuotaUsageRpcRequest(AbstractRpcRequest["AbstractGatewayNode"]):
    help = {
        "params": "",
        "description": "return percentage used of daily transaction quota",
    }

    def validate_params(self) -> None:
        pass

    async def process_request(self) -> JsonRpcResponse:
        if self.node.account_id is not None:
            account_id = self.node.account_id
            assert account_id is not None
            result = sdn_http_service.fetch_quota_status(account_id)
            if result:
                return self.ok(result)
            else:
                raise RpcAccountIdError(
                    self.request_id,
                    "Gateway is not registered with a valid account."
                )
        else:
            raise RpcAccountIdError(
                self.request_id,
                "Gateway is not registered with an account."
            )



# TODO: make these integers to save us some bytes
class BloxrouteMessageType:
    ABSTRACT_INTERNAL = None
    HELLO = b"hello"
    ACK = b"ack"
    PING = b"ping"
    PONG = b"pong"
    BROADCAST = b"broadcast"
    TRANSACTION = b"tx"
    GET_TRANSACTIONS = b"gettxs"
    TRANSACTIONS = b"txs"
    GET_TX_CONTENTS = b"getcontents"
    TX_CONTENTS = b"txcontents"
    KEY = b"key"
    BLOCK_HOLDING = b"blockhold"
    DISCONNECT_RELAY_PEER = b"droprelay"
    TX_SERVICE_SYNC_REQ = b"txstart"
    TX_SERVICE_SYNC_BLOCKS_SHORT_IDS = b"txblock"
    TX_SERVICE_SYNC_TXS = b"txtxs"
    TX_SERVICE_SYNC_COMPLETE = b"txdone"
    BLOCK_CONFIRMATION = b"blkcnfrm"
    TRANSACTION_CLEANUP = b"txclnup"
    NOTIFICATION = b"notify"
    BDN_PERFORMANCE_STATS = b"bdnstats"
    REFRESH_BLOCKCHAIN_NETWORK = b"blkntwrk"
    GET_COMPRESSED_BLOCK_TXS = b"getblocktxs"
    COMPRESSED_BLOCK_TXS = b"blocktxs"
    ROUTING_UPDATE = b"routing"



import struct
from typing import Optional

from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash
from bxgateway import ont_constants
from bxgateway.messages.ont import ont_messages_util
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType


class TxOntMessage(OntMessage):
    MESSAGE_TYPE = OntMessageType.TRANSACTIONS

    def __init__(self, magic: Optional[int] = None, version: Optional[int] = None, txload: Optional[bytes] = None,
                 buf: Optional[bytearray] = None):
        if buf is None:
            assert version is not None
            assert txload is not None
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF + len(txload) + ont_constants.ONT_CHAR_LEN)
            self.buf = buf

            off = ont_constants.ONT_HDR_COMMON_OFF
            struct.pack_into("<B", buf, off, version)
            off += ont_constants.ONT_CHAR_LEN

            #  param but got `Optional[bytes]`.
            buf[off:off + len(txload)] = txload
            off += len(txload)

            super().__init__(magic, self.MESSAGE_TYPE, off - ont_constants.ONT_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._version = self._txload = self._tx_hash = None

    def __repr__(self):
        return ("TxOntMessage<version: {}, length: {}, tx_hash: {}, rawbytes: {}>"
                .format(self.version(),
                        len(self.rawbytes()),
                        convert.bytes_to_hex(self.tx_hash().binary),
                        convert.bytes_to_hex(self.rawbytes().tobytes())
                        )
                )

    def version(self) -> int:
        if self._version is None:
            off = ont_constants.ONT_HDR_COMMON_OFF
            self._version, = struct.unpack_from("<B", self.buf, off)
        version = self._version
        assert isinstance(version, int)
        return version

    def tx(self) -> bytearray:
        return self.payload()

    def tx_hash(self) -> OntObjectHash:
        if self._tx_hash is None:
            self._tx_hash, _ = ont_messages_util.get_txid(self.payload())
        tx_hash = self._tx_hash
        assert isinstance(tx_hash, OntObjectHash)
        return tx_hash
